/** @file
 * @brief	GDB Remote Serial Protocol
 *
 * @date	2002.03.23
 * @author	Takashi SHUDO
 */

#include "stub.h"
#include "str.h"
#include "task.h"
#include "monio.h"
#include "tprintf.h"
#include "section.h"
#include "stub.h"

enum {
	RSPS_WDOL,	//!< $Ԥ
	RSPS_RCV,	//!< ޥɼ
	RSPS_WSUM0,	//!< SUM 1Х
	RSPS_WSUM1	//!< SUM 2Х
} rsp_stat;	//!< ߤΥޥɼ(RSPS_*)

#define MAXCOMBUF	80	//!< ޥɥХåեΥ
#define SUMLEN		2	//!< åΥ

static char combuf[MAXCOMBUF];	//!< ޥɥХåե
static char sumbuf[SUMLEN+1];	//!< åХåե
static char *comptr;		//!< Хåեݥ

/**
 * @brief	ͤ׻
 * @param[in]	str	ʸݥ
 * @param[in]	len	strΥХȿ
 * @return	׻
 */
static unsigned char calc_sum(char *str, int len)
{
	unsigned char sum = 0;
	
	while(len) {
		sum += *str;
		str ++;
		len --;
	}
	
	return sum;
}

/**
 * @brief	ʸʸΥ֤ͤ
 * @param[in]	str	ʸݥ
 * @return	׻
 */
static unsigned char send_str(char *str)
{
	unsigned char sum = 0;
	
	while(*str) {
		stubputc(*str);
		sum += *str;
		str ++;
	}

	return sum;
}

/**
 * @brief	ʸ
 * @param[in]	str	ʸݥ
 */
static void put_str(char *str)
{
	while(*str) {
		stubputc(*str);
		str ++;
	}
}

/**
 * @brief	stub쥹ݥ󥹥ǡ
 * @param[in]	str	ǡ
 */
static void send_responce(char *str)
{
	unsigned char sum = 0;
	char sumstr[3];

	put_str("+$");
	
	sum += send_str(str);
	
	stubputc('#');
	ltohs(sumstr, sum, 2);
	put_str(sumstr);
}

/**
 * @brief	ƥå׼¹ԥʥ
 */
void stub_send_signal(void)
{
	send_responce("S05");
}

/**
 * @brief	֥졼ʥ
 */
void stub_send_expected(void)
{
	unsigned char sum = 0;
	char str[10];
	int i;
	
	put_str("+$");
	
	sum += send_str("T05");

	for(i=0; i<10; i++) {
		// 쥸ֹ
		ltohs(str, i, 1);
		str[1] = ':';
		str[2] = 0;
		sum += send_str(str);

		// 쥸
		ltohs(str, task_reg.data[i], 8);
		str[8] = ';';
		str[9] = 0;
		sum += send_str(str);
	}
	
	stubputc('#');
	ltohs(str, sum, 2);
	put_str(str);
}

/**
 * @brief	Read Registersޥ<br>
 *  		쥸ͤ
 */
static void com_read_registers(void)
{
	unsigned char sum = 0;
	char str[10];
	int i;
	
	put_str("+$");

	for(i=0; i<10; i++) {
		ltohs(str, task_reg.data[i], 8);
		sum += send_str(str);
	}
	
	stubputc('#');
	ltohs(str, sum, 2);
	put_str(str);
}

/**
 * @brief	queryޥɼ¹
 * @param[in]	str	ޥɰʸ
 */
static void com_question(char *str)
{
	if(strcomp(str, "Offsets") == 0) {
		send_responce("Text=0;Data=0;Bss=0");
	} else if(strcomp(str, "C") == 0) {
		send_responce("QC1");	// ???
	} else {
		send_responce("");	// ̤ݡ
	}
}

/**
 * @brief	Write registerޥ
 * @param[in]	str	ޥɰʸ
 */
static void com_write_registerNN(char *str)
{
	task_reg.data[str[0]-'0'] = hstol(&str[2]);

	send_responce("OK");
}

/**
 * @brief	Write registersޥ
 * @param[in]	str	ޥɰʸ
 */
static void com_write_registers(char *str)
{
	int i;
	
	for(i=0; i<10; i++) {
		task_reg.data[i] = hstol(&str[i*8]);
	}

	send_responce("OK");
}

/**
 * @brief	Read memoryޥ
 * @param[in]	str	ޥɰʸ
 */
static void com_read_memory(char *str)
{
	unsigned long add = 0;
	short len = 0;
	unsigned char sum = 0;
	int i;
	char buf[3];

	// ɥ쥹
	while(*str != ',') {
		add = ((add << 4) + h2c(str));
		str ++;
	}
		
	str ++;
	
	// 
	while(*str != 0) {
		len = ((len << 4) + h2c(str));
		str ++;
	}
	
	// ǡ
	put_str("+$");
	
	for(i=0; i<len; i++) {
		ltohs(buf,
		      (unsigned long)(*(unsigned char *)((INTPOINTER)add)),2);
		sum += send_str(buf);
		add ++;
	}

	stubputc('#');
	ltohs(buf, sum, 2);
	put_str(buf);
}

/**
 * @brief	Write memoryޥ
 * @param[in]	str	ޥɰʸ
 */
static void com_write_memory(char *str)
{
	unsigned char *add = 0;
	short len = 0;
	int i;

	// ɥ쥹
	while(*str != ',') {
		add = (unsigned char *)((((INTPOINTER)add) << 4) + h2c(str));
		str ++;
	}

	str ++;
	
	// 
	while(*str != ':') {
		len = ((len << 4) + h2c(str));
		str ++;
	}
	
	str ++;

	// ǡ
	for(i=0; i<len; i++) {

		if((VECT_START <= (INTPOINTER)add) &&
		   ((INTPOINTER)add < VECT_END)) {
			*(add + VVECT_START)=hs2c(str);
		} else {
			*add = hs2c(str);
		}
		
		str += 2;
		add ++;
	}

	send_responce("OK");
}

/**
 * @brief	Write memory(binary)ޥ
 * @param[in]	str	ޥɰʸ
 */
static void com_write_memory_bin(char *str)
{
	unsigned char *add = 0;
	short len = 0;
	int i;

	// ɥ쥹
	while(*str != ',') {
		DPUTC(*str);
		add = (unsigned char *)((((INTPOINTER)add) << 4) + h2c(str));
		str ++;
	}
	DPUTC(*str);

	str ++;
	
	// 
	while(*str != ':') {
		DPUTC(*str);
		len = ((len << 4) + h2c(str));
		str ++;
	}
	DPUTC(*str);
	
	str ++;

	// ǡ
	for(i=0; i<len; i++) {
		unsigned char *vadd = 0;
		
		DPUTC(*str);
		if((VECT_START <= (INTPOINTER)add) &&
		   ((INTPOINTER)add < VECT_END)) {
			vadd = add + VVECT_START;
		} else {
			vadd = add;
		}
		
		switch(*str) {
		case '#':
		case '$':
		case 0x7d:
			str ++;
			*vadd = (*str + 0x20);
			break;
			
		default:
			*vadd = *str;
			break;
		}
	
		str ++;
		add ++;
	}

	send_responce("OK");
}

/**
 * @brief	stub
 * @param[in]	ch	ʸ
 * @return	1:kill !=1:ʳ
 */
short stub_task(char ch)
{
	switch(rsp_stat) {
	case RSPS_WDOL:	// $Ԥ
		if(ch == '$') {
			rsp_stat = RSPS_RCV;
			comptr = combuf;
		}
		break;
		
	case RSPS_RCV:	// ޥɼ
		if(ch == '#') {
			rsp_stat = RSPS_WSUM0;
		} else {
			if(comptr < &combuf[MAXCOMBUF-SUMLEN-1]) {
				*comptr = ch;
				comptr ++;
				*comptr = 0;
			} else {
				// Хåեե
				put_str("-$#00");
				DPRINTF("Buffer full.\n");
				rsp_stat = RSPS_WDOL;
			}
		}
		break;

	case RSPS_WSUM0:	// 1Х
		sumbuf[0] = ch;
		rsp_stat = RSPS_WSUM1;
		break;

	case RSPS_WSUM1:	// 2Х
		{
			unsigned char csum, sum;
			
			// ޥɼλ
			sumbuf[1] = ch;
			sumbuf[2] = 0;
			
			// å
			csum = calc_sum(combuf, comptr-combuf);
			sum = hstol(sumbuf);
			
			if(csum != sum) {
				// बʤ
				put_str("-$#00");
				DPRINTF("Bad check sum.\n");
			} else {
				switch(combuf[0]) {
					/*
					 * 쥸Ϣޥ
					 */
				case 'g': // "Read registers"
					com_read_registers();
					break;
					
				case 'G': // "Write registers"
					com_write_registers(&combuf[1]);
					break;
					
				case 'P': // "Write register nn"
					com_write_registerNN(&combuf[1]);
					break;

					/*
					 * Ϣޥ
					 */
				case 'm': // "Read Memory"
					com_read_memory(&combuf[1]);
					break;
					
				case 'M': // "Write Memory"
					com_write_memory(&combuf[1]);
					break;

					/*
					 * åȾ
					 */ 
				case 'q': // "query"
					com_question(&combuf[1]);
					break;
					
					/*
					 * ץॳȥ
					 */ 
				case 'H': // "Set thread"
					// "Hc-1"Ȥ뤬åɤϣĤ̵ΤǲǤOK.
					send_responce("OK");
					break;
					
				case 's': // "Step"
					send_responce("OK");
					rsp_stat = RSPS_WDOL;	// ΥޥԤ

					step_exec_usertask();
					break;

				case 'c': // "Continue"
					send_responce("OK");
					rsp_stat = RSPS_WDOL;	// ΥޥԤ

					exec_usertask();
					break;

				case '?': // "Last signal"
					stub_send_signal();
					break;

				case 'k': // "Kill"
					stubputc('+');
					return 1;
					break;
					
				case 'X': // "write mem (binary)"
					com_write_memory_bin(&combuf[1]);
					break;
					
				default:
					send_responce("");	// ̤ݡ
					break;
				}
			}
			
			rsp_stat = RSPS_WDOL;	// ΥޥԤ
		}
		break;
	}

	return 0;
}

/**
 * @brief	stubν
 */
void stub_init(void)
{
	rsp_stat = RSPS_WDOL;
}
